@@ -26,6 +26,7 @@
 #include <strings.h>
 #include <popt.h>
 #include <gdk/gdk.h>
+#include <gdk-pixbuf/gdk-pixbuf-io.h>
 #include <libbonobo.h>
 #include <login-helper/login-helper.h>
 #include <gdk/gdkx.h>
@@ -38,6 +39,7 @@
 #include "magnifier-private.h"
 #include "zoom-region.h"
 #include "zoom-region-private.h"
+#include "damage-client.h"
 #include "GNOME_Magnifier.h"
 
 /* if you #define this, don't forget to set MAG_CLIENT_DEBUG env variable */
@@ -133,6 +135,7 @@ gchar* mag_prop_names[MAGNIFIER_CROSSWIRE_COLOR_PROP + 1] = {
 static int _x_error = 0;
 static int fixes_event_base = 0, fixes_error_base;
 static Display *cursor_client_connection;
+static guint    cursor_client_gsource = 0;
 static Magnifier *_this_magnifier = NULL;
 
 static void magnifier_transform_cursor (Magnifier *magnifier);
@@ -141,6 +144,7 @@ static void magnifier_init_window (Magnifier *magnifier, GdkScreen *screen);
 static gboolean magnifier_check_set_struts (Magnifier *magnifier);
 static gboolean magnifier_reset_struts_at_idle (gpointer data);
 static void magnifier_init_window (Magnifier *magnifier, GdkScreen *screen);
+static void magnifier_adjust_source_size (Magnifier *magnifier);
 static gboolean _is_override_redirect = FALSE;
 
 static Window*
@@ -229,7 +233,7 @@ static gboolean
 can_open_display (gchar *display_name)
 {
     Display *d;
-    if (d = XOpenDisplay (display_name))
+    if ((d = XOpenDisplay (display_name)))
     {
 	XCloseDisplay (d);
 	return TRUE;
@@ -316,7 +320,7 @@ magnifier_set_cursor_from_pixbuf (Magnifier *magnifier, GdkPixbuf *cursor_pixbuf
 		gdk_pixbuf_render_threshold_alpha (cursor_pixbuf, mask, 0, 0, 0, 0, 
 						   width, height,
 						   200);
-		gdk_gc_unref (gc);
+		g_object_unref (gc);
 		magnifier->priv->cursor = pixmap;
 		magnifier->priv->cursor_mask = mask;
 		xhot_string = gdk_pixbuf_get_option (cursor_pixbuf,"x_hot");
@@ -413,7 +417,7 @@ magnifier_set_cursor_pixmap_by_name (Magnifier *magnifier, const gchar *cursor_n
 		}
 	}
 	magnifier_set_cursor_from_pixbuf (magnifier, pixbuf);
-	if (pixbuf) gdk_pixbuf_unref (pixbuf);
+	if (pixbuf) g_object_unref (pixbuf);
 }
 
 gboolean
@@ -422,7 +426,7 @@ magnifier_cursor_notify (GIOChannel *source, GIOCondition condition, gpointer da
 #ifdef HAVE_XFIXES
   XEvent ev;
   Magnifier *magnifier = (Magnifier *) data;
-  XFixesCursorNotifyEvent  *cev;
+  XFixesCursorNotifyEvent  *cev = NULL;
 
   /* TODO: don't ask for name unless xfixes version >= 2 available */
   do
@@ -438,17 +442,18 @@ magnifier_cursor_notify (GIOChannel *source, GIOCondition condition, gpointer da
   {
       GdkPixbuf *cursor_pixbuf = magnifier_get_source_pixbuf (magnifier);
       magnifier_set_cursor_from_pixbuf (magnifier, cursor_pixbuf);
-      if (cursor_pixbuf) gdk_pixbuf_unref (cursor_pixbuf);
+      if (cursor_pixbuf) g_object_unref (cursor_pixbuf);
   }
   else
   {
       magnifier_set_cursor_pixmap_by_name (magnifier, 
-					   gdk_x11_get_xatom_name (cev->cursor_name),
+					   cev ? gdk_x11_get_xatom_name (cev->cursor_name) : "default",
 					   TRUE);
   }
 
   magnifier_transform_cursor (magnifier);
 #ifdef DEBUG_CURSOR
+  if (cev)
   g_message ("cursor changed: subtype=%d, cursor_serial=%lu, name=[%x] %s\n",
 	     (int) cev->subtype, cev->cursor_serial, (int) cev->cursor_name,
 	     gdk_x11_get_xatom_name (cev->cursor_name));
@@ -466,13 +471,19 @@ magnifier_cursor_notification_init (Magnifier *magnifier)
 #ifdef HAVE_XFIXES
     GIOChannel *ioc;
     int fd;
-    Display *dpy;
     Window rootwin;
 
     if (!magnifier->source_display) return FALSE;
 
-    dpy = GDK_DISPLAY_XDISPLAY (magnifier->source_display);
-    cursor_client_connection = XOpenDisplay (0);
+    if (cursor_client_connection)
+    {
+        /* remove the old watch */
+	if (cursor_client_gsource) 
+	  g_source_remove (cursor_client_gsource);
+	XCloseDisplay (cursor_client_connection);
+    }
+
+    cursor_client_connection = XOpenDisplay (magnifier->source_display_name);
     rootwin = GDK_WINDOW_XWINDOW (magnifier->priv->root);
 
     if (!XFixesQueryExtension (cursor_client_connection, &fixes_event_base, &fixes_error_base))
@@ -485,8 +496,9 @@ magnifier_cursor_notification_init (Magnifier *magnifier)
 	XFixesSelectCursorInput (cursor_client_connection, rootwin, XFixesDisplayCursorNotifyMask);
 	fd = ConnectionNumber (cursor_client_connection);
 	ioc = g_io_channel_unix_new (fd);
-	g_io_add_watch (ioc, G_IO_IN | G_IO_HUP | G_IO_PRI | G_IO_ERR, magnifier_cursor_notify, 
-			magnifier);
+	cursor_client_gsource = 
+	    g_io_add_watch (ioc, G_IO_IN | G_IO_HUP | G_IO_PRI | G_IO_ERR, magnifier_cursor_notify, 
+			    magnifier);
 	g_io_channel_unref (ioc); 
 	g_message ("added event source to xfixes cursor-notify connection");
  	XFlush (cursor_client_connection); 
@@ -518,7 +530,6 @@ GdkFilterReturn
 magnifier_expose_filter (GdkXEvent *xevent, GdkEvent *event, gpointer data)
 {
 	Magnifier *magnifier = data;
-	GList *list;
 
 	if (event->any.type == GDK_EXPOSE) 
 	{
@@ -662,6 +673,51 @@ magnifier_get_display_rect_bounds (Magnifier *magnifier, GNOME_Magnifier_RectBou
     }
 }
 
+static void magnifier_adjust_source_size (Magnifier *magnifier)
+{
+	GNOME_Magnifier_RectBounds rect_bounds;	
+	magnifier_get_display_rect_bounds (magnifier, &rect_bounds, FALSE);
+	gdouble vfract, hfract;
+	hfract = (double) (magnifier->target_bounds.x2 - magnifier->target_bounds.x1) / (double) (rect_bounds.x2 - rect_bounds.x1);
+	vfract = (double) (magnifier->target_bounds.y2 - magnifier->target_bounds.y1) / (double) (rect_bounds.y2 - rect_bounds.y1);
+	if (vfract > hfract) /* vertical splitpane, approximately */
+	{
+	        gint leftw, rightw;
+		leftw = magnifier->target_bounds.x1 - rect_bounds.x1;
+		rightw = rect_bounds.x2 - magnifier->target_bounds.x2;
+		if (rightw > leftw) {
+			magnifier->source_bounds.x1 = magnifier->target_bounds.x2;
+			magnifier->source_bounds.x2 = rect_bounds.x2;
+		}
+		else 
+		{
+			magnifier->source_bounds.x1 = rect_bounds.x1;
+			magnifier->source_bounds.x2 = magnifier->target_bounds.x1;
+		}
+		magnifier->source_bounds.y1 = rect_bounds.y1;
+		magnifier->source_bounds.y2 = rect_bounds.y2;
+	}
+	else
+	{
+	        gint toph, bottomh;
+		toph = magnifier->target_bounds.y1 - rect_bounds.y1;
+		bottomh = rect_bounds.y2 - magnifier->target_bounds.y2;
+		if (bottomh > toph) {
+			magnifier->source_bounds.y1 = magnifier->target_bounds.y2;
+			magnifier->source_bounds.y2 = rect_bounds.y2;
+		}
+		else 
+		{
+			magnifier->source_bounds.y1 = rect_bounds.y1;
+			magnifier->source_bounds.y2 = magnifier->target_bounds.y1;
+		}
+		magnifier->source_bounds.x1 = rect_bounds.x1;
+		magnifier->source_bounds.x2 = rect_bounds.x2;
+	}
+	g_message ("set source bounds to %d,%d; %d,%d", 
+		   magnifier->source_bounds.x1, magnifier->source_bounds.y1, magnifier->source_bounds.x2, magnifier->source_bounds.y2);
+}
+
 static void
 magnifier_unref_zoom_region (gpointer data, gpointer user_data)
 {
@@ -952,7 +1008,6 @@ magnifier_transform_cursor (Magnifier *magnifier)
 static void
 magnifier_init_cursor_set (Magnifier *magnifier, gchar *cursor_set)
 {
-	GdkPixbuf	*cursor_pixbuf = NULL;
 	/*
 	 * we check the cursor-set property string here,
 	 * and create/apply the appropriate cursor settings
@@ -1125,7 +1180,6 @@ magnifier_get_property (BonoboPropertyBag *bag,
 			gpointer user_data)
 {
 	Magnifier *magnifier = user_data;
-	GNOME_Magnifier_RectBounds rect_bounds;
 	int csize = 0;
 
 	DBG (fprintf (stderr, "Get property: \t%s\n", mag_prop_names[arg_id]));
@@ -1197,7 +1251,6 @@ magnifier_set_property (BonoboPropertyBag *bag,
 			gpointer user_data)
 {
 	Magnifier *magnifier = user_data;
-	GNOME_Magnifier_RectBounds rect_bounds;
 	gchar *full_display_string;
 
 	switch (arg_id) {
@@ -1221,9 +1274,9 @@ magnifier_set_property (BonoboPropertyBag *bag,
 				magnifier->source_screen_num));
 	             /* attach listeners for DAMAGE, "dirty region", XFIXES cursor changes */
 	            magnifier_set_extension_listeners (magnifier, magnifier_get_root (magnifier));
+		    magnifier_get_display_rect_bounds (magnifier, &magnifier->source_bounds, FALSE);
 		    magnifier_zoom_regions_restore (magnifier, zoom_region_params);
 		    magnifier_warp_cursor_to_screen (magnifier);
-		    magnifier_get_display_rect_bounds (magnifier, &magnifier->source_bounds, FALSE);
 		    magnifier_check_set_struts (magnifier);
 		}
 		DBG(fprintf (stderr, "Set source display: \t%s\n", full_display_string));
@@ -1281,7 +1334,8 @@ magnifier_set_property (BonoboPropertyBag *bag,
 		    DBG(fprintf (stderr, "Set target size: \t%d,%d to %d,%d\n", 
 			     magnifier->target_bounds.x1, magnifier->target_bounds.y1, magnifier->target_bounds.x2, magnifier->target_bounds.y2));
 		    }
-		    /* N.B. do this AFTER gtk_window_resize, so as not to change the target */
+		    /* N. B. we don't reset the target bounds to the limits of the new display, because */
+		    /* doing so would override the client-specified magnifier size */
 		    /* magnifier_get_display_rect_bounds (magnifier, &magnifier->target_bounds, TRUE); */
 		    magnifier_check_set_struts (magnifier);
 		}
@@ -1311,6 +1365,9 @@ magnifier_set_property (BonoboPropertyBag *bag,
 		magnifier_check_set_struts (magnifier);
 		DBG(fprintf (stderr, "Set target size: \t%d,%d to %d,%d\n", 
 			      magnifier->target_bounds.x1, magnifier->target_bounds.y1, magnifier->target_bounds.x2, magnifier->target_bounds.y2));
+		if (!strcmp (magnifier->target_display_name, magnifier->source_display_name) && 
+		    (magnifier->target_screen_num == magnifier->source_screen_num)) 
+		    magnifier_adjust_source_size (magnifier);
 		break;
 	case MAGNIFIER_CURSOR_SET_PROP:
 		magnifier_init_cursor_set (magnifier, g_strdup (BONOBO_ARG_GET_STRING (arg)));
