@@ -26,7 +26,7 @@
 #include <sys/time.h> /* TODO: replace timeofday with g_timers */
 #include <gdk/gdkwindow.h>
 #include <gtk/gtk.h>
-#ifndef USE_GDKPIXBUF_RENDER_TO_DRAWABLE
+#ifdef USE_GDKPIXBUF_RENDER_TO_DRAWABLE
 #include <gdk/gdkpixbuf.h>
 #else
 #include <gdk/gdk.h>
@@ -217,6 +217,10 @@ static void zoom_region_paint_pixmap (ZoomRegion *zoom_region, GdkRectangle *rec
 static int  zoom_region_update_pointer_timeout (gpointer data);
 static void zoom_region_recompute_exposed_viewport (ZoomRegion *zoom_region);
 
+static GdkRectangle zoom_region_rect_from_bounds (ZoomRegion *zoom_region,
+						  const GNOME_Magnifier_RectBounds *bounds);
+static ZoomRegionPixmapCreationError zoom_region_create_pixmap (ZoomRegion *zoom_region);
+
 void
 reset_timing_stats()
 {
@@ -597,8 +601,12 @@ static GdkRectangle
 _rectangle_clip_to_rectangle (GdkRectangle area,
 			      GdkRectangle clip_rect)
 {
-	/* TODO: Implement this! */
-	return area;
+        GdkRectangle clipped;
+        clipped.x = MAX (area.x, clip_rect.x);
+        clipped.y = MAX (area.y, clip_rect.y);
+        clipped.width = MIN ((area.x + area.width), (clip_rect.x + clip_rect.width)) - clipped.x;
+        clipped.height = MIN ((area.y + area.height), (clip_rect.y + clip_rect.height)) - clipped.y;
+	return clipped;
 }
 
 static GdkRectangle
@@ -622,6 +630,7 @@ zoom_region_clip_to_source (ZoomRegion *zoom_region,
     if (zoom_region && zoom_region->priv && zoom_region->priv->parent)
     {
 	source_rect_ptr = &((Magnifier *)zoom_region->priv->parent)->source_bounds;
+        DEBUG_RECT ("clipping to source bounds", zoom_region_rect_from_bounds (zoom_region, source_rect_ptr)); 
 	return _rectangle_clip_to_bounds (area, source_rect_ptr);
     }
     return area;
@@ -631,15 +640,17 @@ static GdkRectangle
 zoom_region_clip_to_exposed_target (ZoomRegion *zoom_region,
 				    GdkRectangle area)
 {
-	GNOME_Magnifier_RectBounds onscreen_target;
+	GNOME_Magnifier_RectBounds onscreen_target, *source_area;
+	source_area = &zoom_region->priv->source_area;
+
 	onscreen_target.x1 = MAX (zoom_region->priv->exposed_bounds.x1 / zoom_region->xscale,
-				  zoom_region->priv->source_area.x1);
+				  source_area->x1);
 	onscreen_target.y1 = MAX (zoom_region->priv->exposed_bounds.y1 / zoom_region->yscale,
-				  zoom_region->priv->source_area.y1);
+				  source_area->y1);
 	onscreen_target.x2 = MIN (zoom_region->priv->exposed_bounds.x2 / zoom_region->xscale,
-				  zoom_region->priv->source_area.x2);
+				  source_area->x2);
 	onscreen_target.y2 = MIN (zoom_region->priv->exposed_bounds.y2 / zoom_region->yscale,
-				  zoom_region->priv->source_area.y2);
+				  source_area->y2);
 
 	return _rectangle_clip_to_bounds (area, &onscreen_target);
 }
@@ -660,11 +671,18 @@ zoom_region_clip_to_viewport (ZoomRegion *zoom_region,
 }
 #endif
 
-static const GdkRectangle
-zoom_region_clip_to_target (ZoomRegion *zoom_region,
-			    GdkRectangle area)
+static GdkRectangle
+zoom_region_clip_to_scaled_pixmap (ZoomRegion *zoom_region,
+				   GdkRectangle area)
 {
-	return _rectangle_clip_to_bounds (area, &zoom_region->priv->source_area);
+        GdkRectangle pixmap_area = {0, 0, 0, 0};
+	if (zoom_region->priv && zoom_region->priv->pixmap)
+	{
+	    gdk_drawable_get_size (zoom_region->priv->pixmap, &pixmap_area.width, &pixmap_area.height);
+	    return _rectangle_clip_to_rectangle (area, pixmap_area);
+	}
+	else
+	    return area;
 }
 
 static GdkRectangle
@@ -774,11 +792,17 @@ zoom_region_update_current (ZoomRegion *zoom_region)
 #ifdef ZOOM_REGION_DEBUG
 	g_assert (zoom_region->alive);
 #endif
-	if (zoom_region->priv && GDK_IS_DRAWABLE (zoom_region->priv->pixmap))
-	    zoom_region_update (zoom_region,
-				zoom_region_source_rect_from_view_bounds (
-					zoom_region,
-					&zoom_region->priv->exposed_viewport));
+	if (zoom_region->priv)
+	{
+		gboolean pixmap_valid = GDK_IS_DRAWABLE (zoom_region->priv->pixmap);
+		if (!pixmap_valid)
+			pixmap_valid = (zoom_region_create_pixmap (zoom_region) == ZOOM_REGION_ERROR_NONE);
+		if (pixmap_valid)
+			zoom_region_update (zoom_region,
+					    zoom_region_source_rect_from_view_bounds (
+						    zoom_region,
+						    &zoom_region->priv->exposed_viewport));
+	}
 }
 
 static GdkRectangle
@@ -1068,7 +1092,7 @@ zoom_region_coalesce_updates (ZoomRegion *zoom_region)
 		zoom_region->priv->q = NULL; /* just discard and update everything */
 		/* CAUTION: this can be an expensive operation! */
 		zoom_region_queue_update (zoom_region, zoom_region_rect_from_bounds
-			(zoom_region, &zoom_region->priv->source_area));
+					  (zoom_region, &zoom_region->priv->source_area));
 	}
 	else 
 
@@ -1780,12 +1804,17 @@ zoom_region_update (ZoomRegion *zoom_region,
 	{
 		GdkRectangle paint_rect;
 		gettimeofday(&mag_timing.scale_start, NULL);
+		DEBUG_RECT ("source rect", source_rect);
 		paint_rect = zoom_region_view_rect_from_source_rect (zoom_region, source_rect);
-		paint_rect = zoom_region_clip_to_window (zoom_region, paint_rect);
+		/* paint_rect = zoom_region_clip_to_scaled_pixmap (zoom_region, paint_rect); */
 		DEBUG_RECT ("paint rect", paint_rect);
 		if (zoom_region->invert)
 			_zoom_region_invert_pixbuf (subimage);
 
+		/** 
+		 *   XXX: We seem to be breaking with the original intention here, which was to 
+		 *   keep a fullscreen scaled pixbuf in-sync.
+		 **/
 		gdk_pixbuf_scale (subimage,
 				  zoom_region->priv->scaled_pixbuf,
 				  0,
@@ -2267,6 +2296,9 @@ zoom_region_set_viewport (ZoomRegion *zoom_region,
 	if (!zoom_region->priv->w) {
 		zoom_region_init_window (zoom_region);
 	} else {
+	        CORBA_any *any;
+		CORBA_Environment ev;
+		Bonobo_PropertyBag properties;
 		Magnifier *magnifier = (Magnifier *) zoom_region->priv->parent;
 		GtkFixed *fixed = GTK_FIXED (magnifier->priv->canvas);
 		gtk_fixed_move (fixed,
@@ -2278,6 +2310,20 @@ zoom_region_set_viewport (ZoomRegion *zoom_region,
 					     zoom_region->viewport.x1,
 					     zoom_region->viewport.y2 -
 					     zoom_region->viewport.y1);
+		CORBA_exception_init (&ev);
+		properties = 
+			GNOME_Magnifier_Magnifier_getProperties(
+				BONOBO_OBJREF (
+					(Magnifier *) zoom_region->priv->parent), &ev);
+		if (!BONOBO_EX (&ev))
+			any = Bonobo_PropertyBag_getValue (
+				properties, "source-display-bounds", &ev);
+		if (!BONOBO_EX (&ev))
+			zoom_region->priv->source_area =
+				*((GNOME_Magnifier_RectBounds *) any->_value);
+		if (zoom_region->priv->pixmap) 
+			g_object_unref (zoom_region->priv->pixmap);
+		zoom_region_create_pixmap (zoom_region);
 	}
 	zoom_region_queue_update (zoom_region,
 				  zoom_region_source_rect_from_view_bounds (
@@ -2670,10 +2716,10 @@ impl_zoom_region_set_mag_factor (PortableServer_Servant servant,
 
 	zoom_region->priv->scaled_pixbuf = gdk_pixbuf_new (
 		GDK_COLORSPACE_RGB, FALSE, 8,
-		zoom_region->priv->source_area.x2 -
-		zoom_region->priv->source_area.x1 * zoom_region->xscale + 1,
-		zoom_region->priv->source_area.y2 -
-		zoom_region->priv->source_area.y1 * zoom_region->yscale + 1);
+		(zoom_region->priv->source_area.x2 -
+		zoom_region->priv->source_area.x1) * zoom_region->xscale + 1,
+		(zoom_region->priv->source_area.y2 -
+		zoom_region->priv->source_area.y1) * zoom_region->yscale + 1);
 
 	if (zoom_region->priv->pixmap) {
 		g_object_unref (zoom_region->priv->pixmap);
@@ -2685,12 +2731,13 @@ impl_zoom_region_set_mag_factor (PortableServer_Servant servant,
 		zoom_region_create_pixmap (zoom_region);
 		g_object_unref (zoom_region->priv->scaled_pixbuf);
 
+		/* only create a scaled image big enough for the target display, for now */
 		zoom_region->priv->scaled_pixbuf = gdk_pixbuf_new (
 			GDK_COLORSPACE_RGB, FALSE, 8,
-			zoom_region->priv->source_area.x2 -
-			zoom_region->priv->source_area.x1 * zoom_region->xscale + 1,
-			zoom_region->priv->source_area.y2 -
-			zoom_region->priv->source_area.y1 * zoom_region->yscale + 1);
+			(zoom_region->priv->source_area.x2 -
+			zoom_region->priv->source_area.x1) * zoom_region->xscale + 1,
+			(zoom_region->priv->source_area.y2 -
+			zoom_region->priv->source_area.y1) * zoom_region->yscale + 1);
 		retval = CORBA_FALSE;
 	}
 	zoom_region_update_current (zoom_region);
@@ -2747,7 +2794,7 @@ impl_zoom_region_mark_dirty (PortableServer_Servant servant,
 	zoom_region_update_pointer (zoom_region, TRUE);
 	/* XXX ? should we clip here, or wait till process_updates? */
 	zoom_region_queue_update (zoom_region, 
-	  zoom_region_clip_to_target (zoom_region, 
+	  zoom_region_clip_to_source (zoom_region, 
 	      zoom_region_rect_from_bounds (zoom_region, roi_dirty)));
 }
 
